use serde::Deserialize;
use std::{env,fmt,fs::File};
use std::io::BufReader;
use std::collections::{HashMap, HashSet};
use std::path::{Path, PathBuf};
use std::process::ExitCode;
use anyhow::Result;

mod parsing;
use parsing::*;

// parses a diplonat port parameter: ({tcp,udp}_port XX YY...)
// matches `DiplonatParameter` in diplonat (`src/consul_actor.rs`)
#[derive(Deserialize, Debug)]
enum DiplonatParameter {
    #[serde(rename = "tcp_port")]
    TcpPort(HashSet<u16>),
    #[serde(rename = "udp_port")]
    UdpPort(HashSet<u16>),
}

// parses a diplonat consul tag: (diplonat <param>...)
// matches `DiplonatConsul` in diplonat (`src/consul_actor.rs`)
#[derive(Deserialize, Debug)]
enum DiplonatTag {
    #[serde(rename = "diplonat")]
    Diplonat(Vec<DiplonatParameter>),
}

struct Ctx {
    // Maps datacenter+port to corresponding registrations.
    // If there is more than one registration, then there is a conflict.
    ports: HashMap<(String, Port), Vec<PortRegistration>>,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
enum Port { Udp(u16), Tcp(u16) }

#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
enum PortRegistration {
    Config {
        file: PathBuf,
        job: String,
        group: String,
        task: String,
        service: Option<String>
    },
    Builtin(String),
}

impl fmt::Display for PortRegistration {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            PortRegistration::Config { file, job, group, task, service } =>
                write!(f, "in {}, job \"{}\", group \"{}\", task \"{}\", service \"{}\"",
                       file.display(), job, group, task,
                       match &service {
                           None => "<noname>",
                           Some(s) => s
                       }),
            PortRegistration::Builtin(reason) =>
                write!(f, "{}", reason),
        }
    }
}

impl Ctx {
    pub fn new() -> Self {
        Ctx { ports: HashMap::new() }
    }

    // ingest a whole .hcl file
    pub fn add_file(&mut self, file: &Path) -> anyhow::Result<()> {
        let cfg: Config = {
            let file = File::open(file)?;
            let config: Config = hcl::from_reader(BufReader::new(file))?;
            config
        };

        for (jobname, job) in cfg.jobs {
            for (groupname, group) in job.groups {
                for (taskname, task) in group.tasks {
                    for service in task.services {
                        for tag in service.tags {
                            let diplo_conf: serde_lexpr::error::Result<DiplonatTag> =
                                serde_lexpr::from_str(&tag);
                            if let Ok(conf) = diplo_conf {
                                self.add_diplonat_tag(
                                    conf,
                                    &job.datacenters,
                                    PortRegistration::Config {
                                        file: PathBuf::from(file),
                                        job: jobname.clone(),
                                        group: groupname.clone(),
                                        task: taskname.clone(),
                                        service: service.name.clone(),
                                    }
                                );
                            }
                        }
                    }
                }
            }
        }
        Ok(())
    }

    // helper function: ingest a diplonat tag, of the form "(diplonat (tcp_port XX YY) (udp_port ZZ UU))"
    fn add_diplonat_tag(&mut self, tag: DiplonatTag, datacenters: &[String], reg: PortRegistration) {
        let DiplonatTag::Diplonat(params) = tag;
        for datacenter in datacenters {
            for param in &params {
                match param {
                    DiplonatParameter::TcpPort(ports) => {
                        for &p in ports {
                            self.add_port(Port::Tcp(p), datacenter.clone(), reg.clone())
                        }
                    },
                    DiplonatParameter::UdpPort(ports) => {
                        for &p in ports {
                            self.add_port(Port::Udp(p), datacenter.clone(), reg.clone())
                        }
                    }
                }
            }
        }
    }

    // helper function: insert a port+datacenter+registration info
    fn add_port(&mut self, p: Port, datacenter: String, reg: PortRegistration) {
        let k = (datacenter, p);
        match self.ports.get_mut(&k) {
            None => { self.ports.insert(k, vec![reg]); },
            Some(ports) => ports.push(reg)
        }
    }
}

fn main() -> Result<ExitCode> {
    let args: Vec<_> = env::args().collect();
    if args.len() <= 1 {
        eprintln!("usage: {} <file.hcl>...", args[0]);
        return Ok(ExitCode::FAILURE)
    }

    let mut ctx = Ctx::new();

    // port registrations managed externally and not described in .hcl files
    ctx.ports.insert((String::from("bespin"), Port::Tcp(80)),
                     vec![PortRegistration::Builtin(String::from("forjego"))]);
    ctx.ports.insert((String::from("bespin"), Port::Tcp(443)),
                     vec![PortRegistration::Builtin(String::from("forjego"))]);

    for file in &args[1..] {
        ctx.add_file(Path::new(file))?
    }

    let mut have_conflicts = false;
    for ((datacenter, port), regs) in ctx.ports {
        if regs.len() > 1 {
            have_conflicts = true;
            println!("Conflict in site {}, port {:?}:", &datacenter, port);
            for reg in &regs {
                println!("- {}", reg)
            }
            println!()
        }
    }

    Ok(if have_conflicts { ExitCode::FAILURE } else { ExitCode::SUCCESS })
}
