use serde::de::{
    value::{MapAccessDeserializer, SeqAccessDeserializer},
    MapAccess, SeqAccess, Visitor,
};
use serde::{Deserialize, Deserializer};
use std::fmt;
use std::marker::PhantomData;

#[derive(Deserialize, Debug)]
pub struct Config {
    #[serde(rename = "job")]
    pub jobs: hcl::Map<String, Job>,
}

#[derive(Deserialize, Debug)]
pub struct Job {
    pub datacenters: Vec<String>,
    #[serde(rename = "group", default = "hcl::Map::new")]
    pub groups: hcl::Map<String, Group>,
}

#[derive(Deserialize, Debug)]
pub struct Group {
    #[serde(rename = "task", default = "hcl::Map::new")]
    pub tasks: hcl::Map<String, Task>,
}

#[derive(Deserialize, Debug)]
pub struct Task {
    #[serde(rename = "service", deserialize_with = "vec_or_map", default = "Vec::new")]
    pub services: Vec<Service>,
}

#[derive(Deserialize, Debug)]
pub struct Service {
    pub name: Option<String>,
    #[serde(default = "Vec::new")]
    pub tags: Vec<String>,
}

// In .hcl files, a field can sometimes container either a record (for a single
// item), or a sequence of records (for several items). This is not something
// that the built-in hcl deserializer handles. Instead, the following
// deserializer can be attached to fields that follow this schema with the
// "deserialize_with" serde attribute.
//
// Thanks to the hcl-rs maintener for the code snippet
// (https://github.com/martinohmann/hcl-rs/issues/380#issuecomment-2456546232)
fn vec_or_map<'de, T, D>(deserializer: D) -> Result<Vec<T>, D::Error>
where
    T: Deserialize<'de>,
    D: Deserializer<'de>,
{
    struct VecOrMap<T>(PhantomData<fn() -> T>);

    impl<'de, T> Visitor<'de> for VecOrMap<T>
    where
        T: Deserialize<'de>,
    {
        type Value = Vec<T>;

        fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
            formatter.write_str("sequence or map")
        }

        fn visit_seq<A>(self, seq: A) -> Result<Self::Value, A::Error>
        where
            A: SeqAccess<'de>,
        {
            Deserialize::deserialize(SeqAccessDeserializer::new(seq))
        }

        fn visit_map<M>(self, map: M) -> Result<Self::Value, M::Error>
        where
            M: MapAccess<'de>,
        {
            Deserialize::deserialize(MapAccessDeserializer::new(map)).map(|value| vec![value])
        }
    }

    deserializer.deserialize_any(VecOrMap(PhantomData))
}
